"""Firewall policies as part of a UniFi network."""

from ..models.firewall_policy import FirewallPolicy, FirewallPolicyListRequest
from .api_handlers import APIHandler


class FirewallPolicies(APIHandler[FirewallPolicy]):
    """Represents FirewallPolicies configurations."""

    obj_id_key = "_id"
    item_cls = FirewallPolicy
    api_request = FirewallPolicyListRequest.create()



"""Test firewall zones API.

pytest --cov-report term-missing --cov=aiounifi.firewall_zone tests/test_firewall_zones.py
"""

import pytest

from aiounifi.models.firewall_zone import FirewallZoneUpdateRequest

from .fixtures import FIREWALL_ZONES


@pytest.mark.parametrize("firewall_zone_payload", [FIREWALL_ZONES])
@pytest.mark.usefixtures("_mock_endpoints")
async def test_firewall_zones(unifi_controller, unifi_called_with):
    """Test that we get the expected firewall zones."""
    firewall_zones = unifi_controller.firewall_zones
    await firewall_zones.update()
    assert unifi_called_with("get", "/v2/api/site/default/firewall/zone")
    assert len(firewall_zones.values()) == 2

    zone = firewall_zones["678ccc26e3849d2932432e26"]
    assert zone.id == "678ccc26e3849d2932432e26"
    assert zone.name == "LAN"
    assert zone.attr_no_edit is True
    assert zone.default_zone is True
    assert zone.network_ids == ["678ccc26e3849d2932432e20"]
    assert zone.zone_key == "lan"


@pytest.mark.parametrize("is_unifi_os", [True])
@pytest.mark.usefixtures("_mock_endpoints")
async def test_no_firewall_zones(unifi_controller, unifi_called_with):
    """Test that no firewall zones also work."""
    firewall_zones = unifi_controller.firewall_zones
    await firewall_zones.update()
    assert unifi_called_with("get", "/proxy/network/v2/api/site/default/firewall/zone")
    assert len(firewall_zones.values()) == 0


@pytest.mark.parametrize("is_unifi_os", [True])
async def test_firewall_zone_update_request(
    mock_aioresponse, unifi_controller, unifi_called_with
):
    """Test that firewall zone can be updated."""
    zone = FIREWALL_ZONES[0]
    zone_id = zone["_id"]

    mock_aioresponse.put(
        f"https://host:8443/proxy/network/v2/api/site/default/firewall/zone/{zone_id}",
        payload={},
    )

    await unifi_controller.request(FirewallZoneUpdateRequest.create(zone))

    assert unifi_called_with(
        "put",
        f"/proxy/network/v2/api/site/default/firewall/zone/{zone_id}",
        json=zone,
    )



"""Firewall zones as part of a UniFi network."""

from ..models.firewall_zone import FirewallZone, FirewallZoneListRequest
from .api_handlers import APIHandler


class FirewallZones(APIHandler[FirewallZone]):
    """Represents FirewallZones configurations."""

    obj_id_key = "_id"
    item_cls = FirewallZone
    api_request = FirewallZoneListRequest.create()



"""Device port implementation."""

from typing import cast

from .api import ApiItem
from .device import TypedDevicePortTable


class Port(ApiItem):
    """Represents a network port."""

    raw: TypedDevicePortTable

    @property
    def ifname(self) -> str | None:
        """Port name used by USG."""
        return self.raw.get("ifname")

    @property
    def media(self) -> str | None:
        """Media port is connected to."""
        return self.raw.get("media")

    @property
    def name(self) -> str:
        """Port name."""
        if (name := self.raw["name"]) == "":
            # Unifi controller allows to set an empty port name, but it
            # shows up as "Port N" consistently across UI. We mirror the
            # behavior, as empty name is rarely visually helpful.
            return f"Port {self.port_idx}"
        return name

    @property
    def port_idx(self) -> int | None:
        """Port index."""
        return self.raw.get("port_idx")

    @property
    def poe_caps(self) -> int | None:
        """Port PoE capabilities.

        0 - no caps
        3 - auto (PoE/PoE+)
        35 - auto (Poe/PoE+/PoE++)
        7 - 24V passive
        8 - Passthrough
        """
        return self.raw.get("poe_caps")

    @property
    def poe_class(self) -> str | None:
        """Port PoE class."""
        return self.raw.get("poe_class")

    @property
    def poe_enable(self) -> bool | None:
        """Is PoE supported/requested by client."""
        return self.raw.get("poe_enable")

    @property
    def poe_mode(self) -> str | None:
        """Is PoE auto, pasv24, passthrough, off or None."""
        return self.raw.get("poe_mode")

    @property
    def poe_power(self) -> str | None:
        """Is PoE power usage."""
        return self.raw.get("poe_power")

    @property
    def poe_voltage(self) -> str | None:
        """Is PoE voltage usage."""
        return self.raw.get("poe_voltage")

    @property
    def portconf_id(self) -> str | None:
        """Port configuration ID."""
        return self.raw.get("portconf_id")

    @property
    def port_poe(self) -> bool | None:
        """Is PoE used."""
        return self.raw.get("port_poe")

    @property
    def rx_bytes(self) -> int:
        """Bytes received."""
        return self.raw.get("rx_bytes", 0)

    @property
    def rx_bytes_r(self) -> int:
        """Bytes recently received."""
        return cast(int, self.raw.get("rx_bytes-r", 0))

    @property
    def tx_bytes(self) -> int:
        """Bytes transferred."""
        return self.raw.get("tx_bytes", 0)

    @property
    def tx_bytes_r(self) -> int:
        """Bytes recently transferred."""
        return cast(int, self.raw.get("tx_bytes-r", 0))

    @property
    def up(self) -> bool | None:
        """Is port up."""
        return self.raw.get("up")

    @property
    def enabled(self) -> bool | None:
        """Is port enabled."""
        return self.raw.get("enable")

    def __repr__(self) -> str:
        """Return the representation."""
        return f"<{self.name}: Poe {self.poe_enable}>"



"""Hotspot vouchers as part of a UniFi network."""

from ..models.api import TypedApiResponse
from ..models.voucher import (
    Voucher,
    VoucherCreateRequest,
    VoucherDeleteRequest,
    VoucherListRequest,
)
from .api_handlers import APIHandler


class Vouchers(APIHandler[Voucher]):
    """Represents Hotspot vouchers."""

    obj_id_key = "_id"
    item_cls = Voucher
    api_request = VoucherListRequest.create()

    async def create(self, voucher: Voucher) -> TypedApiResponse:
        """Create voucher on controller."""
        return await self.controller.request(
            VoucherCreateRequest.create(
                quota=voucher.quota,
                expire_number=int(
                    voucher.duration.total_seconds() / 60  # Get minutes.
                ),
                usage_quota=voucher.qos_usage_quota,
                rate_max_up=voucher.qos_rate_max_up,
                rate_max_down=voucher.qos_rate_max_down,
                note=voucher.note,
            )
        )

    async def delete(self, voucher: Voucher) -> TypedApiResponse:
        """Delete voucher from controller."""
        return await self.controller.request(
            VoucherDeleteRequest.create(
                obj_id=voucher.id,
            )
        )



"""Aiounifi errors."""


class AiounifiException(Exception):
    """Base error for aiounifi."""


class RequestError(AiounifiException):
    """Unable to fulfill request.

    Raised when host or API cannot be reached.
    """


class ResponseError(AiounifiException):
    """Invalid response."""


class Unauthorized(AiounifiException):
    """Username is not authorized."""


class LoginRequired(AiounifiException):
    """User is logged out."""


class Forbidden(AiounifiException):
    """Forbidden request."""


class NoPermission(AiounifiException):
    """Users permissions are read only."""


class ServiceUnavailable(RequestError):
    """Service is unavailable.

    Common error if controller is restarting and behind a proxy.
    """


class BadGateway(RequestError):
    """Invalid response from the upstream server."""


class TwoFaTokenRequired(AiounifiException):
    """2 factor authentication token required."""


class WebsocketError(AiounifiException):
    """Websocket error."""



"""Firewall zones as part of a UniFi network."""

from dataclasses import dataclass
from typing import NotRequired, Self, TypedDict

from .api import ApiItem, ApiRequestV2


class TypedFirewallZone(TypedDict):
    """Firewall zone type definition."""

    _id: str
    name: str
    attr_no_edit: bool
    default_zone: bool
    network_ids: list[str]
    zone_key: NotRequired[str]


class FirewallZone(ApiItem):
    """Represent a firewall zone."""

    raw: TypedFirewallZone

    @property
    def id(self) -> str:
        """Unique ID of firewall zone."""
        return self.raw["_id"]

    @property
    def name(self) -> str:
        """Firewall zone name."""
        return self.raw["name"]

    @property
    def attr_no_edit(self) -> bool:
        """Whether zone is editable."""
        return self.raw["attr_no_edit"]

    @property
    def default_zone(self) -> bool:
        """Whether this is a default zone."""
        return self.raw["default_zone"]

    @property
    def network_ids(self) -> list[str]:
        """List of network IDs associated with zone."""
        return self.raw["network_ids"]

    @property
    def zone_key(self) -> str | None:
        """Zone key identifier."""
        return self.raw.get("zone_key")


@dataclass
class FirewallZoneListRequest(ApiRequestV2):
    """Request object for listing firewall zones."""

    @classmethod
    def create(cls) -> Self:
        """Create firewall zone list request."""
        return cls(method="get", path="/firewall/zone", data=None)


@dataclass
class FirewallZoneUpdateRequest(ApiRequestV2):
    """Request object for updating a firewall zone."""

    @classmethod
    def create(cls, zone: TypedFirewallZone) -> Self:
        """Create firewall zone update request."""
        return cls(
            method="put",
            path=f"/firewall/zone/{zone['_id']}",
            data=zone,
        )



"""Test firewall policies API.

pytest --cov-report term-missing --cov=aiounifi.firewall_policy tests/test_firewall_policies.py
"""

import pytest

from aiounifi.models.firewall_policy import FirewallPolicyUpdateRequest

from .fixtures import FIREWALL_POLICIES


@pytest.mark.parametrize("firewall_policy_payload", [FIREWALL_POLICIES])
@pytest.mark.usefixtures("_mock_endpoints")
async def test_firewall_policies(unifi_controller, unifi_called_with):
    """Test that we get the expected firewall policies."""
    firewall_policies = unifi_controller.firewall_policies
    await firewall_policies.update()
    assert unifi_called_with("get", "/v2/api/site/default/firewall-policies")
    assert len(firewall_policies.values()) == 1

    policy = firewall_policies["678ceb9fe3849d293243405c"]
    assert policy.id == "678ceb9fe3849d293243405c"
    assert policy.action == "ALLOW"
    assert policy.connection_state_type == "ALL"
    assert policy.connection_states == []
    assert policy.create_allow_respond is True
    assert policy.description == ""
    assert policy.destination == {
        "match_opposite_ports": False,
        "matching_target": "ANY",
        "port_matching_type": "ANY",
        "zone_id": "678ccc26e3849d2932432e26",
    }
    assert policy.enabled is True
    assert policy.icmp_typename == "ANY"
    assert policy.icmp_v6_typename == "ANY"
    assert policy.index == 10000
    assert policy.ip_version == "BOTH"
    assert policy.logging is False
    assert policy.match_ip_sec is False
    assert policy.match_opposite_protocol is False
    assert policy.name == "Allow internal to IoT"
    assert policy.predefined is False
    assert policy.protocol == "all"
    assert policy.schedule == {
        "mode": "EVERY_DAY",
        "repeat_on_days": [],
        "time_all_day": False,
        "time_range_end": "12:00",
        "time_range_start": "09:00",
    }
    assert policy.source == {
        "match_opposite_ports": False,
        "matching_target": "ANY",
        "port_matching_type": "ANY",
        "zone_id": "678c63bc2d97692f08adcdfa",
    }


@pytest.mark.parametrize("is_unifi_os", [True])
@pytest.mark.usefixtures("_mock_endpoints")
async def test_no_firewall_policies(unifi_controller, unifi_called_with):
    """Test that no firewall policies also work."""
    firewall_policies = unifi_controller.firewall_policies
    await firewall_policies.update()
    assert unifi_called_with(
        "get", "/proxy/network/v2/api/site/default/firewall-policies"
    )
    assert len(firewall_policies.values()) == 0


@pytest.mark.parametrize("is_unifi_os", [True])
async def test_firewall_policy_update_request(
    mock_aioresponse, unifi_controller, unifi_called_with
):
    """Test that firewall policy can be updated."""
    policy = FIREWALL_POLICIES[0]
    policy_id = policy["_id"]

    mock_aioresponse.put(
        f"https://host:8443/proxy/network/v2/api/site/default/firewall-policies/{policy_id}",
        payload={},
    )

    await unifi_controller.request(FirewallPolicyUpdateRequest.create(policy))

    assert unifi_called_with(
        "put",
        f"/proxy/network/v2/api/site/default/firewall-policies/{policy_id}",
        json=policy,
    )



"""Firewall policies as part of a UniFi network."""

from dataclasses import dataclass
from typing import NotRequired, Self, TypedDict

from .api import ApiItem, ApiRequestV2


class FirewallPolicySchedule(TypedDict):
    """Schedule settings for firewall policy."""

    mode: str
    repeat_on_days: list[str]
    time_all_day: bool


class FirewallPolicyEndpoint(TypedDict):
    """Source or destination endpoint configuration."""

    match_opposite_ports: bool
    matching_target: str
    port_matching_type: str
    zone_id: str
    client_macs: NotRequired[list[str]]


class TypedFirewallPolicy(TypedDict):
    """Firewall policy type definition."""

    _id: str
    action: str
    name: str
    enabled: bool
    connection_state_type: str
    connection_states: list[str]
    create_allow_respond: bool
    description: NotRequired[str]
    destination: FirewallPolicyEndpoint
    source: FirewallPolicyEndpoint
    icmp_typename: str
    icmp_v6_typename: str
    index: int
    ip_version: str
    logging: bool
    match_ip_sec: bool
    match_opposite_protocol: bool
    predefined: bool
    protocol: str
    schedule: FirewallPolicySchedule


class FirewallPolicy(ApiItem):
    """Represent a firewall policy."""

    raw: TypedFirewallPolicy

    @property
    def id(self) -> str:
        """Unique id of firewall policy."""
        return self.raw["_id"]

    @property
    def name(self) -> str:
        """Firewall policy name."""
        return self.raw["name"]

    @property
    def enabled(self) -> bool:
        """Is firewall policy enabled."""
        return self.raw["enabled"]

    @property
    def action(self) -> str:
        """Firewall policy action."""
        return self.raw["action"]

    @property
    def predefined(self) -> bool:
        """Is this a predefined policy."""
        return self.raw["predefined"]

    @property
    def description(self) -> str:
        """Policy description."""
        return self.raw.get("description", "")

    @property
    def protocol(self) -> str:
        """Policy protocol."""
        return self.raw["protocol"]

    @property
    def connection_state_type(self) -> str:
        """Connection state type."""
        return self.raw["connection_state_type"]

    @property
    def connection_states(self) -> list[str]:
        """List of connection states."""
        return self.raw["connection_states"]

    @property
    def create_allow_respond(self) -> bool:
        """Whether to create allow respond rule."""
        return self.raw["create_allow_respond"]

    @property
    def destination(self) -> FirewallPolicyEndpoint:
        """Destination endpoint configuration."""
        return self.raw["destination"]

    @property
    def source(self) -> FirewallPolicyEndpoint:
        """Source endpoint configuration."""
        return self.raw["source"]

    @property
    def icmp_typename(self) -> str:
        """ICMP type name."""
        return self.raw["icmp_typename"]

    @property
    def icmp_v6_typename(self) -> str:
        """ICMPv6 type name."""
        return self.raw["icmp_v6_typename"]

    @property
    def index(self) -> int:
        """Policy index/priority."""
        return self.raw["index"]

    @property
    def ip_version(self) -> str:
        """IP version (IPv4/IPv6/BOTH)."""
        return self.raw["ip_version"]

    @property
    def logging(self) -> bool:
        """Whether logging is enabled."""
        return self.raw["logging"]

    @property
    def match_ip_sec(self) -> bool:
        """Whether to match IPSec traffic."""
        return self.raw["match_ip_sec"]

    @property
    def match_opposite_protocol(self) -> bool:
        """Whether to match opposite protocol."""
        return self.raw["match_opposite_protocol"]

    @property
    def schedule(self) -> FirewallPolicySchedule:
        """Policy schedule configuration."""
        return self.raw["schedule"]


@dataclass
class FirewallPolicyListRequest(ApiRequestV2):
    """Request object for listing firewall policies."""

    @classmethod
    def create(cls) -> Self:
        """Create firewall policy list request."""
        return cls(method="get", path="/firewall-policies", data=None)


@dataclass
class FirewallPolicyUpdateRequest(ApiRequestV2):
    """Request object for updating a firewall policy. Note: You can't update the default policies - only ones you've created."""

    @classmethod
    def create(cls, policy: TypedFirewallPolicy) -> Self:
        """Create method for firewall policy update request."""
        return cls(
            method="put",
            path=f"/firewall-policies/{policy['_id']}",
            data=policy,
        )
